package me.loganb1max.minions.manager;

import com.google.gson.*;
import lombok.Getter;
import me.loganb1max.minions.MinionsPlugin;
import me.loganb1max.minions.menu.MinionMenu;
import me.loganb1max.minions.model.Minion;
import me.loganb1max.minions.util.Text;
import org.bukkit.Bukkit;
import org.bukkit.entity.ArmorStand;
import org.bukkit.event.EventHandler;
import org.bukkit.event.EventPriority;
import org.bukkit.event.HandlerList;
import org.bukkit.event.Listener;
import org.bukkit.event.entity.EntityDamageEvent;
import org.bukkit.event.player.PlayerArmorStandManipulateEvent;
import org.bukkit.scheduler.BukkitTask;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.util.HashSet;
import java.util.Objects;
import java.util.Set;

public class MinionManager implements Listener {

	private final MinionsPlugin plugin;
	private final BukkitTask animationTask;
	private final BukkitTask tickTask;

	@Getter private Set<Minion> minions = new HashSet<>();

	public MinionManager(final MinionsPlugin plugin) {
		this.plugin = plugin;
		this.animationTask = Bukkit.getScheduler().runTaskTimer(this.plugin, () -> this.minions.forEach(Minion::animationTick), 10, 10);
		this.tickTask = Bukkit.getScheduler().runTaskTimer(this.plugin, () -> this.minions.forEach(Minion::tick), 20, 20);
		Bukkit.getPluginManager().registerEvents(this, this.plugin);
		loadMinions();
	}

	public void close() {
		this.animationTask.cancel();
		this.tickTask.cancel();
		HandlerList.unregisterAll(this);
		saveMinions();
	}

	private void loadMinions() {
		final File dataFile = new File(this.plugin.getDataFolder(), "minions.json");
		final Gson gson = new Gson();
		try (final BufferedReader reader = Files.newBufferedReader(dataFile.toPath(), StandardCharsets.UTF_8)) {
			JsonObject data = gson.fromJson(reader, JsonObject.class);
			JsonArray minions = data.get("minions").getAsJsonArray();
			for (JsonElement el : minions) {
				try {
					Minion minion = Minion.deserialize(this.plugin, el);
					this.minions.add(minion);
				} catch (Exception e) {
					this.plugin.getLogger().severe("Failed to load minion.");
				}
			}
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	private void saveMinions() {
		final File dataFile = new File(this.plugin.getDataFolder(), "minions.json");
		final Gson gson = new Gson();
		try (final BufferedWriter writer = Files.newBufferedWriter(dataFile.toPath(), StandardCharsets.UTF_8)) {
			JsonObject data = new JsonObject();
			JsonArray minions = new JsonArray();

			for (Minion minion : this.minions) {
				try {
					minions.add(minion.serialize());
				} catch (Exception e) {
					this.plugin.getLogger().severe("Failed to save minion: " + minion.getOwnerName() + " : " + minion.getName());
					e.printStackTrace();
				}
			}

			data.add("minions", minions);
			writer.write(gson.toJson(data));
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	@EventHandler(priority = EventPriority.HIGHEST)
	public void onDamage(final EntityDamageEvent e) {
		if (!(e.getEntity() instanceof ArmorStand)) return;
		if (this.minions.stream().map(Minion::getId).filter(Objects::nonNull).anyMatch(uuid -> uuid.equals(e.getEntity().getUniqueId()))) {
			e.setCancelled(true);
		}
	}

	@EventHandler(priority = EventPriority.HIGHEST)
	public void onRightClick(final PlayerArmorStandManipulateEvent e) {
		this.minions.stream().filter(minion -> minion.getId().equals(e.getRightClicked().getUniqueId())).findFirst().ifPresent(minion -> {
			e.setCancelled(true);
			if (!minion.getOwner().equals(e.getPlayer().getUniqueId())) {
				e.getPlayer().sendMessage(Text.color("&cOnly the owner of the minion can interact with it."));
				return;
			}
			if (e.getPlayer().isSneaking()) {
				minion.despawn(e.getPlayer());
				return;
			}
			new MinionMenu(minion, e.getPlayer());
		});
	}

}
