package me.loganb1max.minions.model;

import com.google.common.collect.ImmutableSet;
import com.google.gson.Gson;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import lombok.AccessLevel;
import lombok.Getter;
import lombok.Setter;
import me.loganb1max.minions.MinionsPlugin;
import me.loganb1max.minions.model.minions.FarmerMinion;
import me.loganb1max.minions.model.minions.FisherMinion;
import me.loganb1max.minions.model.minions.MinerMinion;
import me.loganb1max.minions.util.*;
import org.bukkit.Bukkit;
import org.bukkit.Location;
import org.bukkit.Material;
import org.bukkit.block.Block;
import org.bukkit.configuration.ConfigurationSection;
import org.bukkit.entity.ArmorStand;
import org.bukkit.entity.EntityType;
import org.bukkit.entity.Player;
import org.bukkit.inventory.ItemStack;
import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.Stream;

@Getter
public abstract class Minion {

	private final ConfigurationSection section;
	private final Replacer replacer;
	private final UUID owner;
	private final String ownerName;
	@Setter(AccessLevel.PROTECTED) private Location location;
	@Setter(AccessLevel.PROTECTED) private double energy;
	@Setter(AccessLevel.PROTECTED) private int level;
	@Setter(AccessLevel.PROTECTED)private Set<Block> linkedBlocks;
	private ArmorStand stand;

	@Getter(AccessLevel.NONE)
	private double offset = 0.1;

	public Minion(final ConfigurationSection section, final UUID owner, final String ownerName) {
		this.section = section;
		this.owner = owner;
		this.ownerName = ownerName;
		this.replacer = Replacer.create()
				.add("%level%", String.valueOf(getLevel()))
				.add("%energy%", String.valueOf(getEnergy()))
				.add("%owner%", getOwnerName());
	}

	protected void findStand() {
		if (this.stand != null) return;
		this.location.getWorld().getNearbyEntities(this.location, 0.5, 0.5, 0.5).stream().filter(entity -> entity instanceof ArmorStand && !entity.getCustomName().isEmpty()).findFirst().ifPresent(entity -> {
			this.stand = (ArmorStand) entity;
		});
	}

	public void spawn(final Location location) {
		this.location = location;
		if (this.stand == null) {
			this.stand = (ArmorStand) location.getWorld().spawnEntity(location, EntityType.ARMOR_STAND);
			this.stand.setSmall(true);
			this.stand.setBasePlate(false);
			this.stand.setGravity(false);
			this.stand.setCanPickupItems(false);
			this.stand.setRemoveWhenFarAway(false);
			this.stand.setMarker(false);
			this.stand.setCustomName(Text.color(getName()));
			this.stand.setCustomNameVisible(true);
			this.equipEquipment(this.stand);
			MinionsPlugin.getInstance().getMinionManager().getMinions().add(this);
		} else {
			this.stand.teleport(location);
		}
	}

	public void despawn(final Player player) {
		if (this.stand == null) return;
		this.location = null;
		this.stand.remove();
		this.stand = null;
		this.linkedBlocks.clear();
		player.getInventory().addItem(this.toItemStack()).forEach((integer, itemStack) -> player.getWorld().dropItemNaturally(player.getLocation(), itemStack));
		MinionsPlugin.getInstance().getMinionManager().getMinions().remove(this);
	}

	public void animationTick() {
		if (this.stand == null) return;
		if (this.stand.getRightArmPose().getX() > 300) offset = -0.1;
		if (this.stand.getRightArmPose().getX() < 230) offset = 0.1;
		this.stand.setRightArmPose(this.stand.getRightArmPose().add(offset, 0, 0));
		this.stand.setFireTicks(0);
	}

	public double gainEnergy(final double amount) {
		return this.energy += amount;
	}

	public double consumeEnergy(final double amount) {
		return this.energy -= amount;
	}

	public void equipEquipment(final ArmorStand stand) {
		final String skull = this.section.getString("Equipment.Skull", "");
		ItemStack skullItem = SkullCreator.itemFromBase64(skull);
		stand.setHelmet(skullItem);
		final ItemStack chestplate = ItemUtil.fromConfigSection(this.section.getConfigurationSection("Equipment.Chestplate"));
		stand.setChestplate(chestplate);
		final ItemStack leggings = ItemUtil.fromConfigSection(this.section.getConfigurationSection("Equipment.Leggings"));
		stand.setLeggings(leggings);
		final ItemStack boots = ItemUtil.fromConfigSection(this.section.getConfigurationSection("Equipment.Boots"));
		stand.setBoots(boots);
		final ItemStack rightHand = ItemUtil.fromConfigSection(this.section.getConfigurationSection("Equipment.RightHand"));
		stand.setItemInHand(rightHand);
	}

	public ItemStack toItemStack() {
		ItemStack item = getIcon();
		item = ItemUtil.setKeyString(item, "minion-owner", this.owner.toString());
		item = ItemUtil.setKeyString(item, "minion-owner-name", this.ownerName);
		item = ItemUtil.setKeyInt(item, "minion-level", this.level);
		item = ItemUtil.setKeyDouble(item, "minion-energy", this.energy);
		return item;
	}

	public ItemStack getIcon() {
		final Material material = Material.getMaterial(this.section.getString("Item.Material"));
		final int durability = this.section.getInt("Item.Durability");
		final String skull = this.section.getString("Item.Skull", null);
		final String displayName = this.section.getString("Item.DisplayName");
		final List<String> lore = this.section.getStringList("Item.Lore");
		ItemStack item = ItemBuilder.of(material)
				.durability((short) durability)
				.name(displayName)
				.lore(this.replacer.replace(lore))
				.build();
		if (material == Material.SKULL_ITEM && skull != null) {
			item = SkullCreator.itemWithBase64(item, skull);
		}
		return item;
	}

	public Set<Block> getLinkedBlocks() {
		return ImmutableSet.copyOf(this.linkedBlocks);
	}

	public UUID getId() {
		if (this.stand == null) return null;
		return this.stand.getUniqueId();
	}

	public String getType() {
		return this.section.getName();
	}

	public String getName() {
		return this.section.getString("Name");
	}

	public double getEnergyPerUse(final int level) {
		return this.section.getDouble("Levels." + level + ".EnergyPerAction", 0.0);
	}

	public double getPrice(final int level) {
		return this.section.getDouble("Levels." + level + ".Price", 0.0);
	}

	public JsonElement serialize() {
		final Gson gson = new Gson();
		final JsonObject object = new JsonObject();
		object.addProperty("type", getType());
		object.addProperty("owner", this.owner.toString());
		object.addProperty("ownerName", this.ownerName);
		final JsonObject loc = new JsonObject();
		loc.addProperty("world", this.location.getWorld().getName());
		loc.addProperty("x", this.location.getX());
		loc.addProperty("y", this.location.getY());
		loc.addProperty("z", this.location.getZ());
		object.add("location", loc);
		object.addProperty("energy", this.energy);
		object.addProperty("level", this.level);
		final List<JsonElement> linked = new ArrayList<>();
		this.linkedBlocks.forEach(block -> {
			JsonObject b = new JsonObject();
			b.addProperty("world", block.getWorld().getName());
			b.addProperty("x", block.getX());
			b.addProperty("y", block.getY());
			b.addProperty("z", block.getZ());
			linked.add(b);
		});
		object.add("linkedBlocks", gson.toJsonTree(linked));
		return object;
	}

	public static Minion deserialize(final MinionsPlugin plugin, final JsonElement element) {
		final JsonObject data = element.getAsJsonObject();
		final String type = data.get("type").getAsString();
		final UUID owner = UUID.fromString(data.get("owner").getAsString());
		final String ownerName = data.get("ownerName").getAsString();
		final JsonObject locData = data.get("location").getAsJsonObject();
		final Location location = new Location(Bukkit.getWorld(locData.get("world").getAsString()), locData.get("x").getAsDouble(), locData.get("y").getAsDouble(), locData.get("z").getAsDouble());
		final double energy = data.get("energy").getAsDouble();
		final int level = data.get("level").getAsInt();
		final Set<Block> linkedBlocks = Stream.of(data.get("linkedBlocks").getAsJsonArray())
				.map(JsonElement::getAsJsonObject)
				.map(jsonObject -> Bukkit.getWorld(jsonObject.get("world").getAsString()).getBlockAt(jsonObject.get("x").getAsInt(), jsonObject.get("y").getAsInt(), jsonObject.get("z").getAsInt())).collect(Collectors.toSet());
		Minion toReturn = null;
		switch (type) {
			case "Miner":
				toReturn = new MinerMinion(owner, ownerName, location, energy, level, linkedBlocks);
				break;
			case "Farmer":
				toReturn = new FarmerMinion(owner, ownerName, location, energy, level, linkedBlocks);
				break;
			case "Fisher":
				toReturn = new FisherMinion(owner, ownerName, location, energy, level, linkedBlocks);
				break;
		}
		return toReturn;
	}

	public abstract void tick();

}
